#include <stdio.h>
#include <zlib.h>


char ct[127];

void generate_ct() {

    int i;

    ct[ 'e' ] = 0x00;
    ct[ 't' ] = 0x01;
    ct[ 'i' ] = 0x02;
    ct[ 'a' ] = 0x03;
    ct[ 'n' ] = 0x04;
    ct[ 'm' ] = 0x05;
    ct[ 's' ] = 0x06;
    ct[ 'u' ] = 0x07;
    ct[ 'r' ] = 0x08;
    ct[ 'w' ] = 0x09;
    ct[ 'd' ] = 0x0a;
    ct[ 'o' ] = 0x0b;
    ct[ 'g' ] = 0x0c;
    ct[ 'h' ] = 0x0d;
    ct[ 'k' ] = 0xe0;
    ct[ '-' ] = 0xe1;
    ct[ 'v' ] = 0xe2;
    ct[ 'f' ] = 0xe3;
    ct[ 'l' ] = 0xe4;
    ct[ 'p' ] = 0xe5;
    ct[ 'j' ] = 0xe6;
    ct[ 'b' ] = 0xe7;
    ct[ 'x' ] = 0xe8;
    ct[ 'c' ] = 0xe9;
    ct[ 'z' ] = 0xea;
    ct[ 'q' ] = 0xeb;
    ct[ 'y' ] = 0xec;

}

#define C(b) ct[ (b) ] 

int compress_string_4bit( char* in, char* out ) {

    char temp;
    int part;
    unsigned char *i, *j;
    unsigned char* size;

    i = in;
    j = out;

    size = j;

    //fprintf( stderr, "%d -> ", j );
    part = 0;

    while( *i && *i != '\n' ) {
        
        if( ( C(*i)  & 0xf0 ) == 0 ) {

            if( part == 0 ) 
                *j = C(*i) << 4;
            else {

                *j += C(*i) & 0x0f;
                j++;

            }

            part = ( part ? 0 : 1 );

        }
        else {

            if( part == 0 ) {

                *j = C(*i);
                j++;

            }
            else {

                *j += 0x0e;
                j++;
                *j = C(*i) << 4;

            }

        }

        i++;
        //fprintf( stderr, "." );

    }

    //fprintf( stderr, "%d\n", j );

    if( part == 0 )
        *( j ) = 0xff;
    else {
        *j = ( *j & 0xf0 ) + 0x0f;

    }

    return j - size + 1;
}

void file_write_4bit_compressed( char* string, FILE* file ) {

    int i, size;
    char compressed[100];

    size = compress_string_4bit( string, compressed );

    write( compressed, size, 1, file );

}

void file_write_4bit_compressed_gzipped( char* string, gzFile file ) {

    int i, size;
    char compressed[100];

    size = compress_string_4bit( string, compressed );

    gzwrite( file, compressed, size );

}

void compress_4bit( char* in_file, char* out_file ) {

    FILE* file; 
    FILE* out;
    int i; 
    char buffer[100];
    char last_word[100];
    unsigned char count;

    if( ( file = fopen( in_file, "r" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n", in_file );
        exit(0);

    }

    if( ( out = fopen( out_file, "wb" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        exit(0);

    }

    while( fgets( buffer, 100, file ) ) {

        file_write_4bit_compressed( buffer, out );

    }


    fclose( file );
    fclose( out );

}

void compress_last_word( char* in_file, char* out_file ) {

    FILE* file; 
    FILE* out;
    int i; 
    char buffer[100];
    char last_word[100];
    unsigned char count;

    if( ( file = fopen( in_file, "r" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n", in_file );
        exit(0);

    }

    if( ( out = fopen( out_file, "w" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        exit(0);

    }

    fgets( last_word, 100, file );
    fputs( last_word, out );
    while( fgets( buffer, 100, file ) ) {

        count = 0;
        while( buffer[ count ] == last_word[ count ] )
            count++;


        putc( count, out );
        fputs( &buffer[count], out );

        strcpy( last_word, buffer );

    }
    
    fclose( file );
    fclose( out );
}

void compress_last_word_4bit( char* in_file, char* out_file ) {

    FILE* file; 
    FILE* out;
    int i; 
    char buffer[100];
    char last_word[100];
    unsigned char count;

    if( ( file = fopen( in_file, "r" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n", in_file );
        exit(0);

    }

    if( ( out = fopen( out_file, "wb" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        exit(0);

    }


    fgets( last_word, 100, file );
    file_write_4bit_compressed( last_word, out );
    while( fgets( buffer, 100, file ) ) {

        count = 0;
        while( buffer[ count ] == last_word[ count ] )
            count++;

        fputc( count, out );
        file_write_4bit_compressed( &buffer[count], out );

        strcpy( last_word, buffer );

    }

    fclose( file );
    fclose( out );

}

void compress_last_word_4bit_gzipped( char* in_file, char* out_file ) {

    FILE* file; 
    gzFile out;
    int i; 
    char buffer[100];
    char last_word[100];
    unsigned char count;

    if( ( file = fopen( in_file, "r" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n", in_file );
        exit(0);

    }

    if( ( out = gzopen( out_file, "wb" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        exit(0);

    }


    gzputs( out, "DICTIONARY_COMPRESSED_4BIT\n" );
    fgets( last_word, 100, file );
    file_write_4bit_compressed_gzipped( last_word, out );
    while( fgets( buffer, 100, file ) ) {

        count = 0;
        while( buffer[ count ] == last_word[ count ] )
            count++;

        gzputc( out, count );
        file_write_4bit_compressed_gzipped( &buffer[count], out );

        strcpy( last_word, buffer );

    }

    fclose( file );
    gzclose( out );

}

void compress_last_word_gzipped( char* in_file, char* out_file ) {

    FILE* file; 
    gzFile* out;
    int i; 
    char buffer[100];
    char last_word[100];
    unsigned char count;

    if( ( file = fopen( in_file, "r" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n", in_file );
        exit(0);

    }

    if( ( out = gzopen( out_file, "wb" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        exit(0);

    }

    gzputs( out, "DICTIONARY_COMPRESSED\n" );
    fgets( last_word, 100, file );
    gzputs( out, last_word );
    while( fgets( buffer, 100, file ) ) {

        count = 0;
        while( buffer[ count ] == last_word[ count ] )
            count++;


        gzputc( out, count );
        gzputs( out, &buffer[count] );

        strcpy( last_word, buffer );

    }
    
    fclose( file );
    gzclose( out );
}


int main( int argc, char** argv ) {

    char out_file[100];
    
    if( argc > 1 ) {

        if( strcmp( argv[1], "-4" ) == 0 ) {
        
    generate_ct();
//    compress_4bit( argv[1], "compressed_4bit" ); 

    sprintf( out_file, "%s.cd", argv[2] );
    compress_last_word_4bit_gzipped( argv[2], out_file ); 

        }
        else {

    sprintf( out_file, "%s.cd", argv[1] );
    compress_last_word_gzipped( argv[1], out_file ); 

        }

//    compress_last_word( argv[1], "compressed_last_word" ); 

    }
    else {

        fprintf( stderr, "Usage: %s [-4] <dictionary>\n  -4\t4bit dictionary compression. Only applicable to English dictionaries\n", argv[0] );

    }

}
