#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <stdlib.h>
#include <ctype.h>
#include <locale.h>

#include "dictionary.h"


static char ct[16];
static char ct2[16];

#define C(b) ct[ (b) ] 
#define C2(b) ct2[ (b) ] 

int _compare_char( char c1, char c2 ) {

    char _compare_char1[2] = { 0, '\0' };
    char _compare_char2[2] = { 0, '\0' };
    _compare_char1[0] = c1;
    _compare_char2[0] = c2;

    return strcoll( _compare_char1, _compare_char2 );

}
    


static char* decompress_4bit_word( char* in ) {

    int part;
    unsigned char temp[100];
    unsigned char *i, *j;

    part = 0;
    i = in;
    j = temp;

    while( !  ( part == 0 & ( *i & 0xf0 ) == 0xf0 ) ||  ( part == 1 &&   ( *i & 0x0f ) == 0x0f ) )  {


        if( part == 0 && ( 0xf0 & *i ) != 0xe0 ) {

            *j++ = C( *i >> 4  );
            part = 1;

        }
        else if( part == 1 && ( 0x0f & *i ) != 0x0e ) {
            
            *j++ = C( *i & 0x0f );
            part = 0;
            i++;

        }
        else if( part == 0 && ( 0xf0 & *i ) == 0xe0 ) {

            *j++ = C2(*i & 0x0f);
            i++;
        
        }
        else if( part == 1 && ( 0x0f & *i ) == 0x0e ) {

            i++;
            *j++ = C2( *i >> 4  );
        
        }

    }
    *j = '\0';    

    //fprintf( stderr, "%s\n", temp );
    return strdup( temp );

}

static char* file_read_4bit_decompressed_gzipped(  gzFile file ) {

    unsigned char compressed[100];
    int in; 
    int i = 0;

    while( ( in = gzgetc( file ) ) != EOF ) {

        compressed[i] = in;
        if( ( in & 0xf0 ) == 0xf0  || ( in & 0x0f ) == 0x0f ) {

            return decompress_4bit_word( compressed ) ;

        }
        i++;

    }

    return NULL;

}

void _dictionary_decompression_create_conversion_table() {

    int i;

    ct[ 0x00 ] = 'e'; 
    ct[ 0x01 ] = 't';
    ct[ 0x02 ] = 'i';
    ct[ 0x03 ] = 'a';
    ct[ 0x04 ] = 'n';
    ct[ 0x05 ] = 'm';
    ct[ 0x06 ] = 's';
    ct[ 0x07 ] = 'u';
    ct[ 0x08 ] = 'r';
    ct[ 0x09 ] = 'w';
    ct[ 0x0a ] = 'd';
    ct[ 0x0b ] = 'o';
    ct[ 0x0c ] = 'g';
    ct[ 0x0d ] = 'h';
    ct2[ 0x00 ] = 'k';
    ct2[ 0x01 ] = '-';
    ct2[ 0x02 ] = 'v';
    ct2[ 0x03 ] = 'f';
    ct2[ 0x04 ] = 'l';
    ct2[ 0x05 ] = 'p';
    ct2[ 0x06 ] = 'j';
    ct2[ 0x07 ] = 'b';
    ct2[ 0x08 ] = 'x';
    ct2[ 0x09 ] = 'c';
    ct2[ 0x0a ] = 'z';
    ct2[ 0x0b ] = 'q';
    ct2[ 0x0c ] = 'y';

}


char* _create_normal_word( const char* word ) {

    int j;
    char* data;


    data = strdup( word );

    for( j = 0; data[j]; j++ )
        data[j] = tolower( data[j] );
    
    return data;

}
char* _create_statistical_word( const char* word ) {

    int size, j;
    char* data;

    size = strlen( word );
    data = (char*)  malloc( ( size + 2 ) * sizeof( char ) );

    data[0] = 0;

    for( j = 1; word[j-1]; j++ )
        data[j] = tolower( word[j-1] );
    data[j] = '\0';

    return data;

}

char* _create_compressed_word_precompressed( const char* ending, int count, char start_char ) {

    int size, j;
    char* data;

    data = (char*)  malloc( ( strlen( ending ) + 3 ) * sizeof( char ) );

    data[0] = count;
    data[1] = start_char;

    for( j = 2; ending[j-2]; j++ )
        data[j] = tolower( ending[j-2] );
    data[j] = '\0';

    return data;

}

 char* _create_compressed_word( const char* word, char* last_word ) {

    int size, j, count;
    char* data;

    count = 0;
    while( word[ count ] == last_word[ count ] )
        count++;

    if( count == 0 )
        return _create_compressed_word_precompressed( word, 0, word[0] );
    else
        return _create_compressed_word_precompressed( word, count, last_word[0] );

}       
 



char _get_word_buffer[100];
#define COMPRESSED_WORD_COUNT(node) ( (char*) node->data )[0]
#define COMPRESSED_WORD_STRING(node) ( &( (char*) node->data)[2] )
#define COMPRESSED_WORD_FIRST_LETTER(node) ( (char*) node->data )[1] 
#define STATISTICAL_WORD_FREQUENCY(node) ( ( (char*) node->data )[0] )
#define STATISTICAL_WORD_STRING(node) ( &( (char*) node->data)[1] )

/*
 * Retrives the string at the node _node_. The returned char* shall _not_ be free()'d
 *
 */
char* _dictionary_get_word( Dictionary* dict,  List* word_node ) {

    int i, count, new_count;
    char* string;
    List* node;
  
    if( word_node == NULL )
        return NULL;

    switch( dict->type ) {

        case DICTIONARY_NORMAL:
            return (char*) word_node->data;
        case DICTIONARY_STATISTICAL:
            //fprintf( stderr, "Ut: %s\n", (char*) &( (char*) node->data)[1] );
            return (char*) &( (char*) word_node->data)[1];
/*        case DICTIONARY_COMPRESSED:

            if( word_node->prev == NULL )
                return (char*) dict->list->data;

            if( COMPRESSED_WORD_COUNT( word_node ) == 0 )
                return COMPRESSED_WORD_STRING( word_node );
            else if( COMPRESSED_WORD_COUNT( word_node ) == 1 )
                return &COMPRESSED_WORD_FIRST_LETTER( word_node );

  
            string = g_hash_table_lookup( dict->decompress_hash, word_node );

            if( string != NULL )
                return string;


            _get_word_buffer[0] = COMPRESSED_WORD_FIRST_LETTER( word_node );


            count = COMPRESSED_WORD_COUNT( word_node );

            strcpy( &_get_word_buffer[count], COMPRESSED_WORD_STRING( word_node ) ); 

            //fprintf( stderr, "%d : %s, %s\n", count, _get_word_buffer, COMPRESSED_WORD_STRING( node ) );
            node =word_node->next;
            while( node && count > 1 ) {

                // check if the node is decompressed
                if( node->data == NULL ) {

                    string = g_hash_table_lookup( dict->decompress_hash, word_node );

                    for( i = 1; i <= count; i++ )
                        _get_word_buffer[i] = string[i]; 

                    count = 0;

                }
                else {
                
                    new_count = i = COMPRESSED_WORD_COUNT( node );

                    if( i < count ) {

                        string = COMPRESSED_WORD_STRING( node );
                        while( i < count )
                            _get_word_buffer[i] = string[ i++ - new_count];

                        count = new_count;

                    }

                }

                node = node->prev;

            }

            string = strdup( _get_word_buffer );
            g_hash_table_replace( dict->decompress_hash, word_node, string );

            // remove the data to save space, since the complete word is held in the hash table
            //free( word_node->data );
            //word_node->data = strdup( "" );
            return string;
*/
    }

    return NULL;

}



static void _dictionary_update_lookup_table( Dictionary* dict ) {

    List* node;
    char current_char, new_char;
    char* string;

    if( dict == NULL || dict->list == NULL )
        return;
 
    setlocale( LC_COLLATE, "C" );
    if( dict->type == DICTIONARY_NORMAL ) {

        current_char = ( (char*) dict->list->data )[0];
        dict->lookup_table[ (unsigned char)  current_char ] =  dict->list;


        for( node = dict->list->next; node; node = node->next ) {

            new_char = ( (char*) node->data )[0];

            if( _compare_char( new_char, current_char ) ) {

                current_char = new_char;

                fprintf( stderr, "%c -> %s\n", current_char, node->data );

                dict->lookup_table[ (unsigned char)  current_char] =  node;

            }

        }

    }
    else if( dict->type == DICTIONARY_STATISTICAL ) {

        current_char = ( (char*) dict->list->data )[1];
        dict->lookup_table[ (unsigned char)  current_char ] =  dict->list;
        //g_hash_table_replace( dict->lookup_table, GINT_TO_POINTER( (int) current_char ), dict->list );

        for( node = dict->list->next; node; node = node->next ) {

            new_char = ( (char*) node->data )[1];

            if( new_char != current_char ) {

                current_char = new_char;
                dict->lookup_table[ (unsigned char)  current_char ] =  node;
                //g_hash_table_replace( dict->lookup_table, GINT_TO_POINTER( (int) current_char ), node );

            }

        }

    }
 
/*    else if( dict->type == DICTIONARY_COMPRESSED ) {

        g_hash_table_replace( dict->lookup_table, GINT_TO_POINTER( (int) ( (char*) dict->list->data )[0]  ), dict->list );
        for( node = dict->list->next; node; node = node->next ) {

            if( COMPRESSED_WORD_COUNT( node ) == 0 ) {

                g_hash_table_replace( dict->lookup_table, GINT_TO_POINTER( (int) COMPRESSED_WORD_STRING( node )[0] ), node );

            }

        }

    }
    */

}

void dictionary_destroy( Dictionary* dict ) {

    List* node;
    List* old;
    int i;
   
    if( dict == NULL )
        return;
    
    list_free_all( dict->list );

    if( dict->last_word )
        free( dict->last_word );
   
   
/*
    if( dict->type == DICTIONARY_COMPRESSED ) {

        g_hash_table_foreach( dict->decompress_hash, _free_hash, NULL );
        g_hash_table_destroy( dict->decompress_hash );

    }
*/
}

void _reduce_statistical_word_list( Dictionary* dict ) {

    List* node, *start_node, *delete_node;
    char current_char;

    node = dict->list;
    while( node ) {

        STATISTICAL_WORD_FREQUENCY( node ) = STATISTICAL_WORD_FREQUENCY( node ) >>  2;

        if( STATISTICAL_WORD_FREQUENCY( node ) == 0 ) {

           current_char = STATISTICAL_WORD_STRING( node )[0];
            start_node = dict->lookup_table[ (unsigned char) current_char ];

            if( node == start_node ) {

                if( node->next != NULL && current_char == STATISTICAL_WORD_STRING( node->next )[0] )
                    dict->lookup_table[ (unsigned char) current_char ] = node->next;
                else
                    dict->lookup_table[ (unsigned char) current_char ] = NULL;

            }

            delete_node = node;
            node = node->next;

            dict->list = list_remove_link( dict->list, delete_node );

        }
        else
            node = node->next;

    }

}

static void _add_word_normal( Dictionary* dict, const char* word, int rehash ) {

    List* first_word, *node, *new_node;
    char* new_word;

    new_word = _create_normal_word( word );

    node = dict->lookup_table[ (unsigned char) word[0] ];

    if( node == NULL ) {

        node = list_append( NULL, new_word );
        dict->list = list_concat( dict->list, node );

    }
    else {

        while( node && strcmp( new_word, (char*) _dictionary_get_word( dict, node ) ) > 0 )
            node = node->next;

        if( node == NULL ) {

            node = list_append( NULL, new_word );
            dict->list = list_concat( dict->list, node );

        }
        else {

            if( strcmp( new_word, (char*) _dictionary_get_word( dict, node ) ) != 0 ) {


                if( node->prev ) {

                    dict->list = list_insert_before( dict->list, node, new_word );
                    node = node->prev;
                
                }
                else {

                    dict->list = list_insert_before( dict->list, node, new_word );
                    node = dict->list;

                }

            }

        }
        
    }

    if( rehash ) {

        first_word = dict->lookup_table[ (unsigned char) word[0] ];

        if( first_word == NULL || strcmp( new_word, (char*) first_word->data ) < 0 )
                dict->lookup_table[ (unsigned char)  new_word[0] ] =  node;

    }

}

static void _add_word_compressed( Dictionary* dict, const char* word, int rehash ) {

    List* first_word, *node, *new_node;
    char* new_word, *string;


    node = dict->lookup_table[ (unsigned char) word[0] ];

    if( node == NULL ) {

        if( dict->list == NULL ) {

            new_word = _create_normal_word( word );
            dict->list = list_append( NULL, new_word );

        }
        else {

            node = list_append( NULL, NULL );
            dict->list = list_concat( dict->list, node );
            node->data = _create_compressed_word( _dictionary_get_word( dict, node->prev ), new_word );

        }


    }
    else {

        while( node && strcmp( new_word, (char*) _dictionary_get_word( dict, node ) ) > 0 )
            node = node->next;

        if( node == NULL ) {


            node = list_append( NULL, _create_compressed_word( _dictionary_get_word( dict, node->prev ), new_word ) );
            dict->list = list_concat( dict->list, node );

        }
        else {

            if( strcmp( new_word, (char*) _dictionary_get_word( dict, node ) ) != 0 ) {


                if( node->prev ) {

                    dict->list = list_insert_before( dict->list, node, _create_compressed_word( _dictionary_get_word( dict, node->prev ), new_word ) );
                    node = node->prev;
                
                }
                else {

                    dict->list = list_insert_before( dict->list, node, new_word );

                    string = (char*) node->data;
                    node->data = _create_compressed_word( new_word, string );
                    free( string );

                    node = dict->list;

                }

            }

        }
        
    }

    if( rehash ) {

        first_word = dict->lookup_table[ (unsigned char) word[0] ];

        if( first_word == NULL || strcmp( new_word, (char*) first_word->data ) < 0 )
                dict->lookup_table[ (unsigned char)  new_word[0] ] =  node;

    }

}

static void _add_word_statistical( Dictionary* dict, const char* in_word, int rehash ) {

    List *node, *node2, *start_node;
    char current_char;
    char* data;
    int current_count;
    char* word = _create_normal_word( in_word );

    start_node = dict->lookup_table[ (unsigned char) word[0] ];

    data = _create_statistical_word( word );

    if( start_node == NULL ) {

            node = list_append( NULL, data ); 
            dict->list = list_concat( dict->list, node );

            if( rehash ) {

                dict->lookup_table[ (unsigned char)  word[0] ] =  node;

            }

    }
    else {

        node = start_node;
        while( node && STATISTICAL_WORD_STRING(node)[0] == word[0] && strcmp( STATISTICAL_WORD_STRING( node ), word ) != 0 )
            node = node->next;

        if( node == NULL ) 
            dict->list = list_append( dict->list, data );
        else if( strcmp( STATISTICAL_WORD_STRING(node), word ) == 0 ) {

            if( STATISTICAL_WORD_FREQUENCY( node ) >= STATISTICAL_WORD_FREQUENCY_LIMIT )
                _reduce_statistical_word_list( dict );

            current_count = ++STATISTICAL_WORD_FREQUENCY( node );

            if( node->prev && STATISTICAL_WORD_FREQUENCY( node->prev ) < current_count && STATISTICAL_WORD_STRING( node->prev )[0] == STATISTICAL_WORD_STRING( node )[0] ) {

                node2 = node;
                node = node->prev;
                
                dict->list = list_remove_link( dict->list, node2 );

                while( node && STATISTICAL_WORD_STRING( node )[0] == word[0] && STATISTICAL_WORD_FREQUENCY( node ) < current_count  ) {

                    node = node->prev;

                }

                if( node == NULL ) {

                    dict->list = list_prepend( dict->list, node2->data );

                    if( rehash ) {

                        dict->lookup_table[ (unsigned char)  word[0] ] =  dict->list;

                    }

                }
                else if( node->next ) {

                    dict->list = list_insert_before( dict->list, node->next, node2->data );

                    if( rehash && node->next == start_node ) {

                        dict->lookup_table[ (unsigned char) word[0] ] = node->prev;

                    }


                }
                else {

                    dict->list = list_append( dict->list, node2->data );

                }

                list_free( node2 );

            }

        }
        else {

            dict->list = list_insert_before( dict->list, node, data );

        }

    }

    free( word );

    
}


void dictionary_add_word( Dictionary* dict, const char* word, int rehash ) {


    if( word == NULL || word[0] == '\0' )
        return;
  
    switch( dict->type ) {

        case DICTIONARY_NORMAL:
            _add_word_normal( dict, word, rehash );
            break;
/*        case DICTIONARY_COMPRESSED:
            _add_word_compressed( dict, word, rehash );
            break;
*/
        case DICTIONARY_STATISTICAL:
            _add_word_statistical( dict, word, rehash );
            break;

    }
    
//    dictionary_print( dict );
    
}

void _remove_word_normal( Dictionary* dict, const char* word, int rehash ) {


    List* node, *deleted_node;
    int compare;
   
    node = dict->lookup_table[ (unsigned char) word[0] ]; 
   
    if( node == NULL )
        return;

    if( rehash && strcmp( node->data, word ) == 0 ) {

        if( node->next && ( (char*) node->next->data)[0] == word[0] ) 
            dict->lookup_table[ (unsigned char) word[0] ] = node->next;
        else
            dict->lookup_table[ (unsigned char) word[0] ] = NULL;

    }
    else { 
    
    for( ; node; node = node->next ) {

        compare = strcmp( node->data, word );
        if( compare == 0 ) {

            deleted_node = node;
            dict->list = list_remove_link( dict->list, node );

            free( node->data );
            list_free( node );

            break;

        }
        else if( compare > 0 )
            break;

    }

    }

}

void _remove_word_compressed( Dictionary* dict, const char* word, int rehash ) {

    List* node, *deleted_node;
   
    node = dict->lookup_table[ (unsigned char) word[0] ]; 
   
    if( node == NULL )
        return;
    if( rehash &&  strcmp( _dictionary_get_word( dict, node ), word ) == 0 ) {

        if( node->next && ( _dictionary_get_word( dict, node->next )[0] == word[0] ) ) 
            dict->lookup_table[ (unsigned char) word[0] ] = node->next;
        else
            dict->lookup_table[ (unsigned char) word[0] ] = NULL;

    }
    else {
    
    for( ; node; node = node->next ) {

        if( strcmp( _dictionary_get_word( dict, node ),  word ) == 0 ) {

            deleted_node = node;
            dict->list = list_remove_link( dict->list, node );

            free( node->data );
            list_free( node );

            break;

        }

    }

    }
    
}

void _remove_word_statistical( Dictionary* dict, const char* word, int rehash ) {

    List* node, *deleted_node;
   
    node = dict->lookup_table[ (unsigned char) word[0] ]; 
   
    if( node == NULL )
        return;

    if( rehash && strcmp( STATISTICAL_WORD_STRING( node ), word ) == 0 ) {

        if( node->next && ( (char*) node->next->data)[0] == word[0] ) 
            dict->lookup_table[ (unsigned char) word[0] ] = node->next;
        else
            dict->lookup_table[ (unsigned char) word[0] ] = NULL;

    }
    else {

        for( ; node && STATISTICAL_WORD_STRING( node )[0] == word[0]; node = node->next ) {

            if( strcmp( node->data, word ) == 0 ) {

                deleted_node = node;
                dict->list = list_remove_link( dict->list, node );

                free( node->data );
                list_free( node );

                break;

            }

        }

    }


}

void dictionary_remove_word( Dictionary* dict, const char* word, int rehash ) {

    List* node;
    List* deleted_node;


    if( word == NULL || word[0] == '\0' )
        return;
  
    switch( dict->type ) {

        case DICTIONARY_NORMAL:
            _remove_word_normal( dict, word, rehash );
            break;

/*        case DICTIONARY_COMPRESSED:
            _remove_word_compressed( dict, word, rehash );
            break;
*/
        case DICTIONARY_STATISTICAL:
            _remove_word_statistical( dict, word, rehash );
            break;

    }

}




void dictionary_add_word_list( Dictionary* dictionary, List* list ) {

    List* node;

    for( node = dictionary->list; node; node = node->next )
        dictionary_add_word( dictionary, (char*) node->data, FALSE );

    _dictionary_update_lookup_table( dictionary );

}

void dictionary_remove_word_list( Dictionary* dictionary, List* list ) {

    List* node;

    for( node = dictionary->list; node; node = node->next )
        dictionary_remove_word( dictionary, (char*) node->data, FALSE );

    _dictionary_update_lookup_table( dictionary );

}

void _dictionary_save_normal( Dictionary* dict, gzFile out ) {

    List* node;
    int i; 
    char* word;
    char* last_word;
    unsigned char count;

    last_word = _dictionary_get_word( dict, dict->list );
    gzputs( out, last_word );
    gzputc( out, '\n' );
    for( node = dict->list->next; node; node = node->next ) {
      
        word = _dictionary_get_word( dict, node );

        if( word == NULL )
            continue;

        count = 0;
        while( word[ count ] == last_word[ count ] )
            count++;

        gzputc( out, count );
        gzputs( out, &word[count] );
        gzputc( out, '\n' );

        last_word = word;

    }

}

/*
void _dictionary_save_compressed( Dictionary* dict, gzFile out ) {

    List* node;
    int i;
    char* word;
    char* last_word;
    unsigned char count;

    last_word = _dictionary_get_word( dict, dict->list );
    gzputs( out, last_word );
    gzputc( out, '\n' );
    for( node = dict->list->next; node; node = node->next ) {
     
        // check if the node is already compressed
        if( node->data == NULL ) {

            node->data = _create_compressed_word( _dictionary_get_word( dict, node->prev ), g_hash_table_lookup( dict->decompress_hash, node ) ); 
            gzputc( out, COMPRESSED_WORD_COUNT( node ) );
            gzputs( out, COMPRESSED_WORD_STRING( node ) );
            gzputc( out, '\n' );

            free( node->data );
            node->data = NULL;

        }
        else {
        
            gzputc( out, COMPRESSED_WORD_COUNT( node ) );
            gzputs( out, COMPRESSED_WORD_STRING( node ) );
            gzputc( out, '\n' );

        }

    }
    
}
*/
int _string_compare_statistical( char* a, char* b ) {

    return strcmp( &a[1], &b[1] );

}

void _dictionary_save_statistical( Dictionary* dict, gzFile out, char* filename ) {

    List* node;
    int i; 
    char* word;
    char* last_word;
    unsigned char count;
    char stat_filename[200];
    FILE* stat_file;

//    fprintf( stderr, "\n ************ Fr ***********\n" );
//    dictionary_print( dict );

    dict->list = list_sort( dict->list, _string_compare_statistical );

//    fprintf( stderr, "\n ************ Etter ***********\n" );
//    dictionary_print( dict );
   
    strcpy( stat_filename, filename );
    strcat( stat_filename, ".stat" );

    if( ( stat_file = fopen( stat_filename, "w" ) ) == NULL ) 
        fprintf( stderr, "Couldn't create the statistics file, statistics will _not_ be saved\n" );
    
    last_word = STATISTICAL_WORD_STRING( dict->list );
    gzputs( out, last_word );
    gzputc( out, '\n' );
    if( stat_file )
        fputc( STATISTICAL_WORD_FREQUENCY( dict->list ), stat_file );

    for( node = dict->list->next; node; node = node->next ) {
      
        word = _dictionary_get_word( dict, node );

        if( word == NULL )
            continue;

        count = 0;
        while( word[ count ] == last_word[ count ] )
            count++;

        gzputc( out, count );
        gzputs( out, &word[count] );
        gzputc( out, '\n' );

        if( stat_file )
            fputc( STATISTICAL_WORD_FREQUENCY( node ), stat_file );

        last_word = word;

    }

    if( stat_file )
        fclose( stat_file );
    
}
    
void dictionary_save( Dictionary* dict, char* filename ) {

    gzFile out;

    if( dict == NULL || dict->list == NULL || filename == NULL )
        return;
    
    if( ( out = gzopen( filename, "wb" ) ) == NULL ) {

        fprintf( stderr, "Error opening file %s\n" );
        return; 

    }
  
    gzputs( out, "DICTIONARY_COMPRESSED\n" );

    switch( dict->type ) {

        case DICTIONARY_NORMAL:
            _dictionary_save_normal( dict, out );
            break;

/*        case DICTIONARY_COMPRESSED:
            _dictionary_save_compressed( dict, out );
            break;
*/
        case DICTIONARY_STATISTICAL:
            _dictionary_save_statistical( dict, out, filename );
            break;

    }
    
    gzclose( out );

    
}

int dictionary_has_word( Dictionary* dict, char* word ) {

    List* node;
    int compare;
    char current_char = tolower( word[0] );
  
    node = dict->lookup_table[ (unsigned char) current_char ];
  
    if( node == NULL )
        return 0;
    
    if( dict->type == DICTIONARY_STATISTICAL ) {

        for( ; node && STATISTICAL_WORD_STRING( node )[0] == current_char ; node = node->next ) {

            compare = strcmp( STATISTICAL_WORD_STRING( node ), word );
            if( compare == 0 )
                return 1;

        }

    }
    else {

        for( ; node && _dictionary_get_word( dict, node )[0] == current_char; node = node->next ) {

            compare = strcmp( _dictionary_get_word( dict,  node ), word );

            //fprintf( stderr, "ER '%s' og '%s' like? %d er svaret\n", _dictionary_get_word( dict, node ), word );
            if( compare == 0 )
                return 1;
            else if( compare > 0 )
                return 0;

        }

    }

    return 0;

}

void dictionary_print( Dictionary* dict ) {

    List* node;
   
    fprintf( stderr, "***************** Dictionary content *******************\n" );
    if( dict->type == DICTIONARY_STATISTICAL ) {

    for( node = dict->list; node; node = node->next )
        fprintf( stderr, "%d: %s\n", STATISTICAL_WORD_FREQUENCY( node ),  _dictionary_get_word( dict, node ) );


    }
    else {

    for( node = dict->list; node; node = node->next )
        fprintf( stderr, "%s\n", _dictionary_get_word( dict, node ) );

    }

}

Dictionary* dictionary_new( dictionary_type dict_type ) {

    Dictionary* dict;
    int i;
    
    dict = (Dictionary*) malloc( sizeof( Dictionary ) );
    dict->list = NULL;
    dict->last_word = strdup( "" );
    dict->last_node = NULL;
    dict->type = dict_type;

    for( i = 0; i < 256; i++ )
        dict->lookup_table[i] = NULL;
    
/*
    if( dict->type == DICTIONARY_COMPRESSED )
        dict->decompress_hash = g_hash_table_new( g_direct_hash, g_direct_equal );
    else
    dict->decompress_hash = NULL;

   */ 


    return dict;

}


Dictionary* dictionary_new_from_file( dictionary_type dict_type, char* filename ) {

    Dictionary* dict = dictionary_new( dict_type );

    dictionary_load( dict, filename );

    return dict;

}

Dictionary* dictionary_new_from_list( dictionary_type dict_type, List* list ) {

    Dictionary* dict = dictionary_new( dict_type );

    dictionary_add_word_list( dict, list );
    return dict;

}

void _dictionary_load_ascii( Dictionary* dict, char* filename ) {

    FILE* file;
    gzFile stat_file = NULL;
    char current_char, lookup_index;
    int i, size, count;
    char buffer[200];
    char last_word[200];
    char stat_filename[110];
    int has_stat = 0;
    char last_char;
    char* data;
    List *node, *last_node;


    fprintf( stderr, "Loading plain dictionary %s\n", filename );

    if( ( file = fopen( filename, "r" ) ) == NULL ) {

        sprintf( buffer, "Error loading dictioary %s: ", filename );
        perror( buffer );
        return;

    }

    if( dict->type == DICTIONARY_STATISTICAL ) {
        
        strcpy( stat_filename, filename );
        strcat( stat_filename, ".stat" );
       
        if( ( stat_file = fopen( stat_filename, "rb" ) ) != NULL )
            has_stat = 1;

    }
   
    fgets( last_word, 100, file );
    last_word[ strlen( last_word ) -1 ] = '\0';
    if( dict->type == DICTIONARY_STATISTICAL ) {

        data = _create_statistical_word( last_word );

        if( stat_file != NULL )
            data[0] = getc( stat_file );

        dict->list = list_prepend( dict->list, data );

        last_node = dict->list;
        last_char = last_word[0];

    }
    else
        dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );


    while( fgets( buffer, 200, file ) ) {

        buffer[ strlen( buffer ) -1 ] = '\0';
        if( dict->type == DICTIONARY_STATISTICAL ) {

            data = _create_statistical_word( buffer );

            if( stat_file != NULL )
                data[0] = getc( stat_file );

            if( buffer[0] != last_char ) {

                last_node = list_append( NULL, data );
                dict->list = list_concat( dict->list, last_node );
                last_char = last_word[0];

            }
            else {

                node = last_node;

                while( node && STATISTICAL_WORD_FREQUENCY( node ) > data[0] )
                    node = node->next;

                if( node == NULL )
                    dict->list = list_append( dict->list, data );
                else
                    dict->list = list_insert_before( dict->list, node, data );

                if( node == last_node )
                    last_node = last_node->prev;

            }
            
       }
/*        else if( dict->type == DICTIONARY_COMPRESSED ) {

            data = _create_compressed_word( buffer, last_word );
            dict->list = list_prepend( dict->list, data );

            strcpy( last_word, buffer );

        }
        */
        else
            dict->list = list_prepend( dict->list, _create_normal_word( buffer ) );
        
    }

    if( dict->type != DICTIONARY_STATISTICAL )
        dict->list = list_reverse( dict->list );
//    else
//        dictionary_print( dict );

    _dictionary_update_lookup_table( dict );


    if( stat_file )
        gzclose( stat_file );

    fclose( file );

} 

void _dictionary_load_gzipped( Dictionary* dict, char* filename ) {

    FILE* file;
    gzFile stat_file = NULL;
    char stat_filename[110];
    char current_char, lookup_index;
    int i, size, count;
    int has_stat = 0;
    char last_char;
    char buffer[100];
    char last_word[100];
    char* data;
    List *node, *last_node;

    fprintf( stderr, "Loading gzipped dictionary %s\n", filename );

    if( ( file = gzopen( filename, "r" ) ) == NULL ) {

        sprintf( buffer, "Error loading dictionary %s: ", filename );
        perror( buffer );
        return;

    }

    if( dict->type == DICTIONARY_STATISTICAL ) {
        
        strcpy( stat_filename, filename );
        strcat( stat_filename, ".stat" );
        
        if( ( stat_file = fopen( stat_filename, "rb" ) ) != NULL )
            has_stat = 1;

    }
    
    gzgets( file, last_word, 100 );
    last_word[ strlen( last_word ) -1 ] = '\0';

    if( dict->type == DICTIONARY_STATISTICAL ) {

        data = _create_statistical_word( last_word );

        if( stat_file != NULL )
            data[0] = getc( stat_file );

        dict->list = list_prepend( dict->list, data );

        last_node = dict->list;
        last_char = last_word[0];

    }
    else
        dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );

    while( gzgets( file, buffer, 100 ) ) {

        if( dict->type == DICTIONARY_STATISTICAL ) {

            data = _create_statistical_word( buffer );

            if( stat_file != NULL )
                data[0] = getc( stat_file );

            if( buffer[0] != last_char ) {

                last_node = list_append( NULL, data );
                dict->list = list_concat( dict->list, last_node );
                last_char = last_word[0];

            }
            else {

                node = last_node;

                while( node && STATISTICAL_WORD_FREQUENCY( node ) > data[0] )
                    node = node->next;

                if( node == NULL )
                    dict->list = list_append( dict->list, data );
                else
                    dict->list = list_insert_before( dict->list, node, data );

                if( node == last_node )
                    last_node = last_node->prev;

            }
            
       }
/*        else if( dict->type == DICTIONARY_COMPRESSED ) {

            data = _create_compressed_word( buffer, last_word );
            dict->list = list_prepend( dict->list, data );

            strcpy( last_word, buffer );

        }
        */
        else
            dict->list = list_prepend( dict->list, _create_normal_word( buffer ) );

    }

    if( dict->type != DICTIONARY_STATISTICAL )
        dict->list = list_reverse( dict->list );
//    else
//        dictionary_print( dict );

    _dictionary_update_lookup_table( dict );

    if( stat_file )
        gzclose( stat_file );

    gzclose( file );


} 

void _dictionary_load_compressed_4bit_gzipped( Dictionary* dict, char* filename ) {

    gzFile file; 
    FILE* stat_file = NULL;
    char stat_filename[110];
    int count, j, size;
    int has_stat = 0;
    char buffer[100];
    char last_word[100];
    char* i;
    char* b;
    char* word;
    char* data;
    List *node, *last_node;
    char last_char;


    fprintf( stderr, "Loading 4bit compressed dictionary %s\n", filename );
    
    if( ( file = gzopen( filename, "rb" ) ) == NULL ) {

        sprintf( buffer, "Error loading dictioary %s: ", filename );
        perror( buffer );
        return;

    }
    
    if( gzgets( file, buffer, 100 ) == NULL ) {

        return;

    }

    if( strcmp( buffer, "DICTIONARY_COMPRESSED_4BIT\n" ) != 0 ) {

        fprintf( stderr, "Error! %s is not a 4bit compressed dictionary!\n", filename );
        return;

    }

    if( dict->type == DICTIONARY_STATISTICAL ) {
        
        strcpy( stat_filename, filename );
        strcat( stat_filename, ".stat" );
        

        if( ( stat_file = fopen( stat_filename, "rb" ) ) != NULL )
            has_stat = 1;

    }
 
    _dictionary_decompression_create_conversion_table();

    word = file_read_4bit_decompressed_gzipped( file ); 
    strcpy( last_word, word );
    free( word );

    if( dict->type == DICTIONARY_STATISTICAL ) {

        data = _create_statistical_word( last_word );

        if( stat_file != NULL )
            data[0] = getc( stat_file );

        dict->list = list_prepend( dict->list, data );

        last_node = dict->list;
        last_char = last_word[0];
        

    }
    else
        dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );

    while( ( count = gzgetc( file ) ) != EOF ) {

        word = file_read_4bit_decompressed_gzipped( file ); 
        
        i = &last_word[count];
        b = word;

        while( *b )
            *i++ = *b++;
        *i = '\0';
       
        if( dict->type == DICTIONARY_STATISTICAL ) {

            data = _create_statistical_word( last_word );

            if( stat_file != NULL )
                data[0] = getc( stat_file );

            if( last_word[0] != last_char ) {

                last_node = list_append( NULL, data );
                dict->list = list_concat( dict->list, last_node );
                last_char = last_word[0];

            }
            else {

                node = last_node;

                while( node && STATISTICAL_WORD_FREQUENCY( node ) > data[0] )
                    node = node->next;

                if( node == NULL )
                    dict->list = list_append( dict->list, data );
                else
                    dict->list = list_insert_before( dict->list, node, data );

                if( node == last_node )
                    last_node = last_node->prev;

            }
           
        }
/*        else if( dict->type == DICTIONARY_COMPRESSED ) {

            data = _create_compressed_word_precompressed( word, count, last_word[0] );
            dict->list = list_prepend( dict->list, data );

        }*/
        else 
            dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );

        free( word );

    }
    
    if( dict->type != DICTIONARY_STATISTICAL )
        dict->list = list_reverse( dict->list );
//    else
//        dictionary_print( dict );


    _dictionary_update_lookup_table( dict );

    gzclose( file );

    if( stat_file )
        gzclose( stat_file );

}

void _dictionary_load_compressed_gzipped( Dictionary* dict, char* filename ) {

    gzFile file; 
    FILE* stat_file = NULL;
    char stat_filename[110];
    int has_stat = 0;
    int count;
    int j, size;
    char buffer[101];
    char last_word[100];
    char* i;
    char* b;
    char* data;
    char last_char;
    List *last_node, *node;

    fprintf( stderr, "Loading compressed dictionary %s\n", filename );
    
    if( ( file = gzopen( filename, "rb" ) ) == NULL ) {

        sprintf( buffer, "Error loading dictioary %s: ", filename );
        perror( buffer );
        return;

    }

    
    if( gzgets( file, buffer, 100 ) == NULL ) {

        return;

    }

    if( strcmp( buffer, "DICTIONARY_COMPRESSED\n" ) != 0 ) {

        fprintf( stderr, "Error! %s is not a compressed dictionary!\n", filename );
        return;

    }

    if( dict->type == DICTIONARY_STATISTICAL ) {
        
        strcpy( stat_filename, filename );
        strcat( stat_filename, ".stat" );
        
        if( ( stat_file = fopen( stat_filename, "rb" ) ) != NULL )
            has_stat = 1;

    }
 
    
    gzgets( file, last_word, 100 );
    last_word[ strlen( last_word ) -1 ] = '\0';

    if( dict->type == DICTIONARY_STATISTICAL ) {

        data = _create_statistical_word( last_word );

        if( stat_file != NULL )
            data[0] = getc( stat_file );

        dict->list = list_prepend( dict->list, data );

        last_node = dict->list;
        last_char = last_word[0];

    }
    else
        dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );

    while( ( count = gzgetc( file ) ) != EOF ) {

        gzgets( file, buffer, 100 );
        buffer[ strlen( buffer ) -1 ] = '\0';
        i = &last_word[count];
        b = buffer;

        while( *b )
            *i++ = *b++;
        *i = '\0';
    
        if( dict->type == DICTIONARY_STATISTICAL ) {

            data = _create_statistical_word( last_word );

            if( stat_file != NULL )
                data[0] = getc( stat_file );

            if( last_word[0] != last_char ) {

                last_node = list_append( NULL, data );
                dict->list = list_concat( dict->list, last_node );
                last_char = last_word[0];

            }
            else {

                node = last_node;

                while( node && STATISTICAL_WORD_FREQUENCY( node ) > data[0] )
                    node = node->next;

                if( node == NULL )
                    dict->list = list_append( dict->list, data );
                else
                    dict->list = list_insert_before( dict->list, node, data );

                if( node == last_node )
                    last_node = last_node->prev;

            }
            
       }
/*        else if( dict->type == DICTIONARY_COMPRESSED ) {

            data = _create_compressed_word_precompressed( buffer, count, last_word[0] );
            dict->list = list_prepend( dict->list, data );

        }*/
        else
            dict->list = list_prepend( dict->list, _create_normal_word( last_word ) );

    }
   
    if( dict->type != DICTIONARY_STATISTICAL )
        dict->list = list_reverse( dict->list );
//    else
//        dictionary_print( dict );

    _dictionary_update_lookup_table( dict );

    if( stat_file )
        fclose( stat_file );
    gzclose( file );

}

void dictionary_load( Dictionary* dict, char* filename ) {

    FILE* file;
    unsigned char check1, check2;
    char buffer[100];

    if( dict == NULL )
        return;
    
    if( ( file = fopen( filename, "r" ) ) == NULL ) {

        sprintf( buffer, "Error loading dictioary %s: ", filename );
        perror( buffer );
        return;

    }

    check1 = fgetc( file );
    check2 = fgetc( file );

    fclose( file );

    // check for the gzip signature
    if( check1 == 0x1f && check2 == 0x8b ) {

        if( ( file = gzopen( filename, "r" ) ) == NULL ) {

            sprintf( buffer, "Error loading dictioary %s: ", filename );
            perror( buffer );
            return;

        }

        if( gzgets( file, buffer, 100 ) == NULL ) {

            gzclose( file );
            return;

        }

        if( strcmp( buffer, "DICTIONARY_COMPRESSED\n" ) == 0 )
           _dictionary_load_compressed_gzipped( dict, filename );
        else if( strcmp( buffer, "DICTIONARY_COMPRESSED_4BIT\n" ) == 0 )
           _dictionary_load_compressed_4bit_gzipped( dict, filename );
        else
            _dictionary_load_gzipped( dict, filename );

    }
    else
        _dictionary_load_ascii( dict, filename );
 
    
}

void dictionary_predict_reset( Dictionary* dict ) {

    if( dict == NULL )
        return;
    
    free( dict->last_word );
    dict->last_word = strdup( "" );
    dict->last_node = NULL;


}

char* dictionary_predict_word( Dictionary* dict, char* word, int cycle_words ) {

    List* node;
    int i, j, k, last_pos, current_pos, old_pos, size;
    char current_char;
    char* current_word;

    if( dict == NULL || dict->list == NULL ) 
        return NULL;

    if(  word == NULL || word[0] == '\0' ) {

        free( dict->last_word );
        dict->last_word = strdup( "" );
        dict->last_node = NULL;

        return NULL;

    }

    current_char = tolower( word[0] );
    node = dict->lookup_table[ (unsigned char) current_char ];

    if( node == NULL ) 
        return NULL;

    size = strlen( word );
    last_pos = 0;

    //fprintf( stderr, "%s == %s, %d->%d\n", word, dict->last_word, dict->last_node, ( dict->last_node != NULL ? dict->last_node->next : -1 ) );
    
    if( cycle_words  && dict->last_word && strcmp( dict->last_word, word ) == 0 && dict->last_node && dict->last_node->next) {

        if( dict->type == DICTIONARY_STATISTICAL ) {

            node = dict->last_node->next;
            dict->last_node = node;

        }
        else {

        current_word = _dictionary_get_word( dict, dict->last_node->next );

        if( strncmp( current_word, word, size ) == 0 ) {

            free( dict->last_word );
            dict->last_word = _create_normal_word( word );

            dict->last_node = dict->last_node->next;
            return current_word;

        }

        }

    }
    else {

        if( dict->last_node && dict->last_word[0] != '\0' && strncmp( dict->last_word, word, strlen( dict->last_word ) ) == 0 ) {
            
            node = dict->last_node;

        }

    }
    
    free( dict->last_word );
    dict->last_word = _create_normal_word( word );
   
    while( node && node->data ) {

        //fprintf( stderr, "%s\n", node->data );
        current_word = _dictionary_get_word( dict, node );

        //fprintf( stderr, "%c != %s == %s\n", current_char, current_word, word );
        if( current_word[0] != current_char )
            break;
       
        //fprintf( stderr, "Fortsatt samme bokstav..\n" );
        for( j = 0; j < size && current_word[j] == dict->last_word[j]; j++ )
            ;

        //fprintf( stderr, "%s = %s, %d == %d?\n", current_word, dict->last_word, j, size );
        if( j == size ) {

            dict->last_node = node;
            return current_word;

        }
        else if( dict->type != DICTIONARY_STATISTICAL && (  j < last_pos || current_word[j] > dict->last_word[j] ) ) {

            //fprintf( stderr, "Hopper av her (%d == %d)..\n", dict->type, DICTIONARY_STATISTICAL );
            break;

        }
        else
            last_pos = j;
  
        //fprintf( stderr, "Lete etter mer....\n" );
        node = node->next;

    }
    
    free( dict->last_word );
    dict->last_word = strdup(""); 
    dict->last_node = NULL;

    return NULL;

}
