#include <stdlib.h>

#include "list.h"

List* list_append( List* list, char* data ) {

    List* new_node, *node;

    new_node = (List*) malloc( sizeof( List ) );
    new_node->data = data;

    if( list == NULL ) {

        new_node->next = new_node->prev = NULL;

        return new_node;

    }
    
    for( node = list; node; node = node->next ) {

        if( node->next == NULL ) {

            node->next = new_node;
            new_node->prev = node;
            new_node->next = NULL;

            break;

        }

    }

    return list;

}

List* list_prepend( List* list, char* data ) {
 
    List* new_node;
    
    new_node = (List*) malloc( sizeof( List ) );
    new_node->data = data;

    new_node->prev = NULL;
    new_node->next = list;

    if( list )
        list->prev = new_node;

    return new_node;

}

   
List* list_concat( List* list, List* concat_list ) {

    List* node;

    if( list == NULL )
        return concat_list;
    
    for( node = list; node; node = node->next ) {

        if( node->next == NULL ) {

            node->next = concat_list;

            if( concat_list )
                concat_list->prev = node;

            break;

        }

    }

    return list;

}


List* list_free( List* list ) {

    List* node, *delete_node;

    for( node = list; node;  ) {

        delete_node = node;
        node = node->next;
        free( delete_node );
        
    }

    return NULL;

}


List* list_free_all( List* list ) {

    List* node, *delete_node;

    for( node = list; node;  ) {

        delete_node = node;
        node = node->next;
        
        if( delete_node->data )
            free( delete_node->data );
        
        free( delete_node );
        
    }

    return NULL;

}


List* list_insert_before( List* list, List* insert_node, char* data ) {

    List* node, *new_node;

    if( list == NULL )
        return NULL;
    
    if( insert_node == list ) 
        return list_prepend( list, data );

    
    for( node = list->next; node; node = node->next ) {

        if( node == insert_node ) {

            new_node = (List*) malloc( sizeof( List ) );
            new_node->data = data;

            new_node->next = node;
            new_node->prev = node->prev;
            node->prev->next = new_node;
            node->prev = new_node;

        }
        
    }

    return list;

}


List* list_remove_link( List* list, List* remove_node ) {

    List* node, *next_node;

    if( list == NULL )
        return NULL;
   
    if( remove_node == list ) {

        node = list->next;
        remove_node->next= remove_node->prev = NULL;
        return node;

    }
    
    for( node = list; node; node = node->next ) {

        if( node == remove_node ) {

            if( node->prev )
                node->prev->next = node->next;
            if( node->next )
                node->next->prev = node->prev;

            node->next = node->prev = NULL;

            break;

        }
        
    }

    return list;

}

List* list_reverse( List* list ) {
    
    List* prev_node, *node;
   
    if( list == NULL )
        return NULL;
    
    for( node = list; ; node = node->prev  ) {

        prev_node = node->prev;
        node->prev = node->next;
        node->next = prev_node;

        if( node->prev == NULL )
            return node;

    }

    return NULL;
    
}

List* list_sort( List* list, int (*compare_func)( char* a, char* b)  ) {

    char* data;
    List* node, *node2;

    if( list == NULL )
        return NULL;
    
    for( node = list->next; node; node = node->next ) {

        data = node->data;

        for( node2 = node; node2->prev && (*compare_func)( data, node2->prev->data ) < 0; node2 = node2->prev )
            node2->data = node2->prev->data;

       node2->data = data;

    }
    
    return list;

}

